# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: dexterity.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0f\x64\x65xterity.proto\x12\tdexterity\"\xa7\x01\n\tMatrix4x4\x12\x0b\n\x03m00\x18\x01 \x01(\x02\x12\x0b\n\x03m01\x18\x02 \x01(\x02\x12\x0b\n\x03m02\x18\x03 \x01(\x02\x12\x0b\n\x03m03\x18\x04 \x01(\x02\x12\x0b\n\x03m10\x18\x05 \x01(\x02\x12\x0b\n\x03m11\x18\x06 \x01(\x02\x12\x0b\n\x03m12\x18\x07 \x01(\x02\x12\x0b\n\x03m13\x18\x08 \x01(\x02\x12\x0b\n\x03m20\x18\t \x01(\x02\x12\x0b\n\x03m21\x18\n \x01(\x02\x12\x0b\n\x03m22\x18\x0b \x01(\x02\x12\x0b\n\x03m23\x18\x0c \x01(\x02\"7\n\x08Skeleton\x12+\n\rjointMatrices\x18\x01 \x03(\x0b\x32\x14.dexterity.Matrix4x4\"X\n\x04Hand\x12)\n\x0bwristMatrix\x18\x01 \x01(\x0b\x32\x14.dexterity.Matrix4x4\x12%\n\x08skeleton\x18\x02 \x01(\x0b\x32\x13.dexterity.Skeleton\"y\n\nHandUpdate\x12\"\n\tleft_hand\x18\x01 \x01(\x0b\x32\x0f.dexterity.Hand\x12#\n\nright_hand\x18\x02 \x01(\x0b\x32\x0f.dexterity.Hand\x12\"\n\x04Head\x18\x03 \x01(\x0b\x32\x14.dexterity.Matrix4x4\"6\n\x0bStereoImage\x12\x12\n\nleft_image\x18\x01 \x01(\x0c\x12\x13\n\x0bright_image\x18\x02 \x01(\x0c\"\x88\x01\n\tSimStates\x12\x34\n\x08matrices\x18\x01 \x03(\x0b\x32\".dexterity.SimStates.MatricesEntry\x1a\x45\n\rMatricesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.dexterity.Matrix4x4:\x02\x38\x01\"\x14\n\x12HealthCheckRequest\"\x8f\x01\n\x13HealthCheckResponse\x12<\n\x06status\x18\x01 \x01(\x0e\x32,.dexterity.HealthCheckResponse.ServingStatus\":\n\rServingStatus\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07SERVING\x10\x01\x12\x0f\n\x0bNOT_SERVING\x10\x02\x32\x90\x02\n\x13HandTrackingService\x12U\n StreamHandUpdatesandGetSimStates\x12\x15.dexterity.HandUpdate\x1a\x14.dexterity.SimStates\"\x00(\x01\x30\x01\x12Z\n#StreamHandUpdatesandGetStereoImages\x12\x15.dexterity.HandUpdate\x1a\x16.dexterity.StereoImage\"\x00(\x01\x30\x01\x12\x46\n\x05\x43heck\x12\x1d.dexterity.HealthCheckRequest\x1a\x1e.dexterity.HealthCheckResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'dexterity_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SIMSTATES_MATRICESENTRY._options = None
  _SIMSTATES_MATRICESENTRY._serialized_options = b'8\001'
  _globals['_MATRIX4X4']._serialized_start=31
  _globals['_MATRIX4X4']._serialized_end=198
  _globals['_SKELETON']._serialized_start=200
  _globals['_SKELETON']._serialized_end=255
  _globals['_HAND']._serialized_start=257
  _globals['_HAND']._serialized_end=345
  _globals['_HANDUPDATE']._serialized_start=347
  _globals['_HANDUPDATE']._serialized_end=468
  _globals['_STEREOIMAGE']._serialized_start=470
  _globals['_STEREOIMAGE']._serialized_end=524
  _globals['_SIMSTATES']._serialized_start=527
  _globals['_SIMSTATES']._serialized_end=663
  _globals['_SIMSTATES_MATRICESENTRY']._serialized_start=594
  _globals['_SIMSTATES_MATRICESENTRY']._serialized_end=663
  _globals['_HEALTHCHECKREQUEST']._serialized_start=665
  _globals['_HEALTHCHECKREQUEST']._serialized_end=685
  _globals['_HEALTHCHECKRESPONSE']._serialized_start=688
  _globals['_HEALTHCHECKRESPONSE']._serialized_end=831
  _globals['_HEALTHCHECKRESPONSE_SERVINGSTATUS']._serialized_start=773
  _globals['_HEALTHCHECKRESPONSE_SERVINGSTATUS']._serialized_end=831
  _globals['_HANDTRACKINGSERVICE']._serialized_start=834
  _globals['_HANDTRACKINGSERVICE']._serialized_end=1106
# @@protoc_insertion_point(module_scope)
