# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: handtracking.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12handtracking.proto\x12\x0chandtracking\"\xdb\x01\n\tMatrix4x4\x12\x0b\n\x03m00\x18\x01 \x01(\x02\x12\x0b\n\x03m01\x18\x02 \x01(\x02\x12\x0b\n\x03m02\x18\x03 \x01(\x02\x12\x0b\n\x03m03\x18\x04 \x01(\x02\x12\x0b\n\x03m10\x18\x05 \x01(\x02\x12\x0b\n\x03m11\x18\x06 \x01(\x02\x12\x0b\n\x03m12\x18\x07 \x01(\x02\x12\x0b\n\x03m13\x18\x08 \x01(\x02\x12\x0b\n\x03m20\x18\t \x01(\x02\x12\x0b\n\x03m21\x18\n \x01(\x02\x12\x0b\n\x03m22\x18\x0b \x01(\x02\x12\x0b\n\x03m23\x18\x0c \x01(\x02\x12\x0b\n\x03m30\x18\r \x01(\x02\x12\x0b\n\x03m31\x18\x0e \x01(\x02\x12\x0b\n\x03m32\x18\x0f \x01(\x02\x12\x0b\n\x03m33\x18\x10 \x01(\x02\"\xb7\x01\n\x0e\x43ontrolCommand\x12\x39\n\x07\x63ommand\x18\x01 \x01(\x0e\x32(.handtracking.ControlCommand.CommandType\x12\x17\n\x0f\x61\x64\x64itional_info\x18\x02 \x01(\t\"Q\n\x0b\x43ommandType\x12\x13\n\x0fSTART_RECORDING\x10\x00\x12\x12\n\x0eSTOP_RECORDING\x10\x01\x12\x08\n\x04SAVE\x10\x02\x12\x0f\n\x0bRESET_SCENE\x10\x03\"3\n\x0f\x43ontrolResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\":\n\x08Skeleton\x12.\n\rjointMatrices\x18\x01 \x03(\x0b\x32\x17.handtracking.Matrix4x4\"^\n\x04Hand\x12,\n\x0bwristMatrix\x18\x01 \x01(\x0b\x32\x17.handtracking.Matrix4x4\x12(\n\x08skeleton\x18\x02 \x01(\x0b\x32\x16.handtracking.Skeleton\"\x82\x01\n\nHandUpdate\x12%\n\tleft_hand\x18\x01 \x01(\x0b\x32\x12.handtracking.Hand\x12&\n\nright_hand\x18\x02 \x01(\x0b\x32\x12.handtracking.Hand\x12%\n\x04Head\x18\x03 \x01(\x0b\x32\x17.handtracking.Matrix4x4\" \n\rHandUpdateAck\x12\x0f\n\x07message\x18\x01 \x01(\t\"$\n\x05\x46rame\x12\x0c\n\x04left\x18\x01 \x01(\x0c\x12\r\n\x05right\x18\x02 \x01(\x0c\"\x1c\n\tStreamAck\x12\x0f\n\x07message\x18\x01 \x01(\t\"\x8c\x01\n\x08Matrices\x12\x36\n\x08matrices\x18\x01 \x03(\x0b\x32$.handtracking.Matrices.MatricesEntry\x1aH\n\rMatricesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12&\n\x05value\x18\x02 \x01(\x0b\x32\x17.handtracking.Matrix4x4:\x02\x38\x01\x32\x62\n\x13HandTrackingService\x12K\n\x11StreamHandUpdates\x12\x18.handtracking.HandUpdate\x1a\x18.handtracking.HandUpdate\"\x00\x30\x01\x32N\n\x0bVideoStream\x12?\n\x0cSendMatrices\x12\x16.handtracking.Matrices\x1a\x17.handtracking.StreamAck2e\n\x0e\x43ontrolService\x12S\n\x12SendControlCommand\x12\x1c.handtracking.ControlCommand\x1a\x1d.handtracking.ControlResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'handtracking_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MATRICES_MATRICESENTRY._options = None
  _MATRICES_MATRICESENTRY._serialized_options = b'8\001'
  _globals['_MATRIX4X4']._serialized_start=37
  _globals['_MATRIX4X4']._serialized_end=256
  _globals['_CONTROLCOMMAND']._serialized_start=259
  _globals['_CONTROLCOMMAND']._serialized_end=442
  _globals['_CONTROLCOMMAND_COMMANDTYPE']._serialized_start=361
  _globals['_CONTROLCOMMAND_COMMANDTYPE']._serialized_end=442
  _globals['_CONTROLRESPONSE']._serialized_start=444
  _globals['_CONTROLRESPONSE']._serialized_end=495
  _globals['_SKELETON']._serialized_start=497
  _globals['_SKELETON']._serialized_end=555
  _globals['_HAND']._serialized_start=557
  _globals['_HAND']._serialized_end=651
  _globals['_HANDUPDATE']._serialized_start=654
  _globals['_HANDUPDATE']._serialized_end=784
  _globals['_HANDUPDATEACK']._serialized_start=786
  _globals['_HANDUPDATEACK']._serialized_end=818
  _globals['_FRAME']._serialized_start=820
  _globals['_FRAME']._serialized_end=856
  _globals['_STREAMACK']._serialized_start=858
  _globals['_STREAMACK']._serialized_end=886
  _globals['_MATRICES']._serialized_start=889
  _globals['_MATRICES']._serialized_end=1029
  _globals['_MATRICES_MATRICESENTRY']._serialized_start=957
  _globals['_MATRICES_MATRICESENTRY']._serialized_end=1029
  _globals['_HANDTRACKINGSERVICE']._serialized_start=1031
  _globals['_HANDTRACKINGSERVICE']._serialized_end=1129
  _globals['_VIDEOSTREAM']._serialized_start=1131
  _globals['_VIDEOSTREAM']._serialized_end=1209
  _globals['_CONTROLSERVICE']._serialized_start=1211
  _globals['_CONTROLSERVICE']._serialized_end=1312
# @@protoc_insertion_point(module_scope)
