# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import dart_physics.avp_stream3.grpc_msg.handtracking_pb2 as handtracking__pb2


class HandTrackingServiceStub(object):
    """The hand tracking service definition.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.StreamHandUpdates = channel.unary_stream(
                '/handtracking.HandTrackingService/StreamHandUpdates',
                request_serializer=handtracking__pb2.HandUpdate.SerializeToString,
                response_deserializer=handtracking__pb2.HandUpdate.FromString,
                )


class HandTrackingServiceServicer(object):
    """The hand tracking service definition.
    """

    def StreamHandUpdates(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_HandTrackingServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'StreamHandUpdates': grpc.unary_stream_rpc_method_handler(
                    servicer.StreamHandUpdates,
                    request_deserializer=handtracking__pb2.HandUpdate.FromString,
                    response_serializer=handtracking__pb2.HandUpdate.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'handtracking.HandTrackingService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class HandTrackingService(object):
    """The hand tracking service definition.
    """

    @staticmethod
    def StreamHandUpdates(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/handtracking.HandTrackingService/StreamHandUpdates',
            handtracking__pb2.HandUpdate.SerializeToString,
            handtracking__pb2.HandUpdate.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)


class VideoStreamStub(object):
    """The video stream service definition
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.SendMatrices = channel.unary_unary(
                '/handtracking.VideoStream/SendMatrices',
                request_serializer=handtracking__pb2.Matrices.SerializeToString,
                response_deserializer=handtracking__pb2.StreamAck.FromString,
                )


class VideoStreamServicer(object):
    """The video stream service definition
    """

    def SendMatrices(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_VideoStreamServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'SendMatrices': grpc.unary_unary_rpc_method_handler(
                    servicer.SendMatrices,
                    request_deserializer=handtracking__pb2.Matrices.FromString,
                    response_serializer=handtracking__pb2.StreamAck.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'handtracking.VideoStream', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class VideoStream(object):
    """The video stream service definition
    """

    @staticmethod
    def SendMatrices(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/handtracking.VideoStream/SendMatrices',
            handtracking__pb2.Matrices.SerializeToString,
            handtracking__pb2.StreamAck.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)


class ControlServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.SendControlCommand = channel.unary_unary(
                '/handtracking.ControlService/SendControlCommand',
                request_serializer=handtracking__pb2.ControlCommand.SerializeToString,
                response_deserializer=handtracking__pb2.ControlResponse.FromString,
                )


class ControlServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def SendControlCommand(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ControlServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'SendControlCommand': grpc.unary_unary_rpc_method_handler(
                    servicer.SendControlCommand,
                    request_deserializer=handtracking__pb2.ControlCommand.FromString,
                    response_serializer=handtracking__pb2.ControlResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'handtracking.ControlService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ControlService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def SendControlCommand(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/handtracking.ControlService/SendControlCommand',
            handtracking__pb2.ControlCommand.SerializeToString,
            handtracking__pb2.ControlResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)
